//
//  psignal_fix.c
//  psignal-fix
//
//  Created by Thomas Tempelmann on 07.01.15.
//
//  See http://blog.tempel.org/2015/01/os-x-kernel-hacking.html
//

#include <mach/mach_types.h>
#include <libkern/libkern.h>
#include <sys/malloc.h>
#include <sys/systm.h>

#include "kernel_info.h"
#include "cpu_protections.h"

#include "distorm.h"
#include "mnemonics.h"


static uint8_t g_patched_area[12];
static void *g_patched_ptr = NULL;

static struct kernel_info g_kernel_info;

/*
 * The code we're going to jump into from the patched code:
 */
void patch_code()
{
	// lead-in are 8 bytes that we want to skip
	current_thread ();
	__asm__ __volatile__ (
		"mov    %rax,%rdx							\n"
		"xor    %edi,%edi                           \n"
		"xor    %esi,%esi                           \n"
		"mov    $0x4,%ecx                           \n"
		"mov    $0x1b,%r8d                          \n"
		"mov    $0x123456789ABCDEF0, %rax           \n"
		"jmp    *%rax                               \n"
	);
}


/*
 * Find the instruction "psignal(p, SIGPROF)" by disassembling bsd_ast()
 * See http://research.swtch.com/macpprof
 */
static boolean_t patch_psignal_SIGPROF()
{
	mach_vm_address_t _bsd_ast = solve_kernel_symbol(&g_kernel_info, "_bsd_ast");
	
	// the instrs we want to find and patch are in 10.9 and earlier:
	//	mov    %r15,%rdi
	//	xor    %esi,%esi
	//	xor    %edx,%edx
	//	xor    %ecx,%ecx
	//	mov    $0x1b,%r8d
	//	callq  <threadsignal+224>
	// and in 10.10:
	//	xor    esi, esi
	//	xor    edx, edx
	//	xor    ecx, ecx
	//	mov    $0x1b,r8d
	//	mov    r15,rdi
	//	callq  ...

	mach_vm_address_t patch_start = 0;
	mach_vm_address_t patch_return = 0;

	#if DEBUG
		printf("[psignal-fix] bst_ast: %llx\n", _bsd_ast);
	#endif

	if (!_bsd_ast) {
		return 0;
	}
	
	// let's disassemble
	#define MAX_INSTRUCTIONS 512
	// allocate space for disassembly output
	_DInst *instr = _MALLOC(sizeof(_DInst) * MAX_INSTRUCTIONS, M_TEMP, M_WAITOK);
	if (instr == NULL) {
		printf("[psignal-fix] ERROR: Decoded instructions allocation failed!\n");
		return 0;
	}
	unsigned int instrCount = 0;
	_DecodeResult res = 0;
	_CodeInfo ci;
	ci.dt = Decode64Bits;
	ci.features = DF_NONE;
	ci.codeLen = MAX_INSTRUCTIONS * 4; // function is not near this big
	ci.code = (uint8_t*)_bsd_ast;
	ci.codeOffset = _bsd_ast;
	mach_vm_address_t next;
	
	while (ci.codeLen >= 0) {
		res = distorm_decompose(&ci, instr, MAX_INSTRUCTIONS, &instrCount);
		if (res == DECRES_INPUTERR) {
			printf("[psignal-fix] ERROR: Distorm failed to disassemble!\n");
			goto failure;
		}

		// iterate over the disassembly and lookup for CALL instructions
		// we're lazy here and do not account for the case where the 6 instrs we're looking
		// for would be only partially disassembled - we expect them all to be in the same
		// decomposed segment, by making (hopefully) MAX_INSTRUCTIONS high enough.
		for (int i = 0; i < instrCount; i++) {
			// look for "mov $0x1b, rax" first
			if ((instr[i].opcode == I_MOV) && (instr[i].ops[0].type == O_REG) && (instr[i].ops[1].type == O_IMM) && (instr[i].imm.ex.i1 == 0x1b)) {
				// found one. now check if the other instrs are there, too
				if (instr[i-1].opcode == I_XOR && instr[i-2].opcode == I_XOR && instr[i-3].opcode == I_XOR) {
					int startofs;
					if (instr[i-4].opcode == I_MOV && instr[i+1].opcode == I_CALL) {
						startofs = i-4;
					} else if (instr[i+1].opcode == I_MOV && instr[i+2].opcode == I_CALL) {
						startofs = i-3;
					} else {
						// not here
						continue;
					}
					#if DEBUG
						for (int j = 0; j <= 5; ++j) {
							printf("[psignal-fix] %d, %d, %d, #%d\n", instr[i+j].opcode, instr[i+j].ops[0].type, instr[i+j].ops[1].type, instr[i+j].imm.ex.i1);
						}
					#endif
					patch_start = instr[startofs].addr;
					patch_return = instr[startofs+5].addr;
					#if DEBUG
						printf("[psignal-fix] found! start %llx, return %llx\n", patch_start, patch_return);
					#endif
					goto success;
				}
			}
		}
		
		if (res == DECRES_SUCCESS || instrCount == 0) {
			// All instructions were decoded.
			#if DEBUG
				printf("[psignal-fix] instructions not found\n");
			#endif
			goto failure;
		}
		
		// sync the disassembly
		// the total number of bytes disassembly to previous last instruction
		next = instr[instrCount-1].addr - ci.codeOffset;
		// add points to the first byte so add instruction size to it
		next += instr[instrCount-1].size;
		// update the CodeInfo struct with the synced data
		ci.code += next;
		ci.codeOffset += next;
		ci.codeLen -= next;
	}

success:
	_FREE(instr, M_TEMP);

	mach_vm_address_t patch_code_ptr = (mach_vm_address_t) (void*) patch_code;
	patch_code_ptr += 8; // skip the function's lead-in code (I learned that by looking at the code with Hopper)

	if (((uint8_t*)patch_code_ptr)[0] != 0xe8) { // call instruction?
		printf("[psignal-fix] ERROR: patch_code not layed out as expected\n");
		return FALSE;
	}
	
	if (*(uint64_t*)(patch_code_ptr+0x19) != 0x123456789ABCDEF0) { // return addr
		printf("[psignal-fix] ERROR: patch_code not layed out as expected (2)\n");
		return FALSE;
	}

	#if DEBUG
		printf("[psignal-fix] patch_code_ptr: %llx\n", patch_code_ptr);
	#endif

	disable_interrupts();
	disable_wp();

	uint8_t *destptr;

	// write the return address into patch_code's push instructions
	destptr = (void*) (patch_code_ptr + 0x19);
	*(uint64_t*)destptr = patch_return;

	// write the mov and jmp instructions to the found area for jumping to the patch code
	//	mov    $0x123456789ABCDEF0, %rax
	//	jmp    %rax
	destptr = (void*) patch_start;
	g_patched_ptr = destptr;
	memcpy(g_patched_area, g_patched_ptr, sizeof(g_patched_area));	// save the original code
	*destptr++ = 0x48; *destptr++ = 0xb8;	// mov
	*(uint64_t*)destptr = patch_code_ptr;	// address
	destptr += 8;
	*destptr++ = 0xFF; *destptr++ = 0xE0;	// jmp rax

	enable_wp();
	enable_interrupts();

	#if DEBUG
		printf("[psignal-fix] patch...\n");
		destptr = (void*) patch_start;
		for (int i = 0; i < 26; ++i) {
			printf(" %02x", (int)*destptr++);
		}
		printf("\n");

		printf("[psignal-fix] new code...\n");
		destptr = (void*) patch_code_ptr;
		for (int i = 0; i < 35; ++i) {
			printf(" %02x", (int)*destptr++);
		}
		printf("\n");

		printf("[psignal-fix] patch_code OK\n");
	#endif
	
	return TRUE;

failure:
	_FREE(instr, M_TEMP);
	return FALSE;
}


static void cleanup()
{
	disable_interrupts();
	disable_wp();
	
	if (g_patched_ptr) {
		memcpy(g_patched_ptr, g_patched_area, sizeof(g_patched_area));	// restore the original code
	}
	
	enable_wp();
	enable_interrupts();

    cleanup_kernel_info(&g_kernel_info);
}

#pragma mark - Termination Handler

extern kern_return_t psignal_fix_stop(kmod_info_t *ki, void *d)
{
	#if DEBUG
		printf("[psignal-fix] Terminating...\n");
	#endif
	
	cleanup();

	printf("[psignal-fix] Terminated\n");

	return KERN_SUCCESS;
}


#pragma mark - Installation Handler

extern kern_return_t psignal_fix_start(kmod_info_t * ki, void *d)
{
	#if DEBUG
		printf("[psignal-fix] Installing...\n");
	#endif

	// Retrieve some essential kernel locations
    if (init_kernel_info(&g_kernel_info) != KERN_SUCCESS) {
		printf("[psignal-fix] Installation failed\n");
		return KERN_FAILURE;
	}
	
	// Patch the psignal code
	if (!patch_psignal_SIGPROF()) {
		cleanup();
		printf("[psignal-fix] Installation failed\n");
		return KERN_FAILURE;
	}

	printf("[psignal-fix] Installed\n");

	return KERN_SUCCESS;
}
